"""
functions and object for managing OSM maps
"""
#!/usr/bin/python
#pylint: skip-file

# import official python packages
import sys
import math
import logging
import geojson
from shapely.geometry import Polygon, shape

# import custom python packages
from wahoomc.constants import GEOFABRIK_PATH
from wahoomc.constants import special_regions, geofabrik_regions, block_download

log = logging.getLogger('main-logger')


class Geofabrik:
    """
    This is a Geofabrik processing class
    """

    def __init__(self, country):
        # input parameters
        self.wanted_map = country
        # replace spaces in self.wanted_map with geofabrik minuses
        self.wanted_map = self.wanted_map.replace(" ", "-")
        self.wanted_map = self.wanted_map.lower()

        self.tiles = []
        self.border_countries = {}
        self.country_name = ''

    def get_tiles_of_country(self):
        """
        Get the relevant tiles for a country
        """

        # Check if wanted_map is in the json file and if so get the polygon (shape)
        wanted_map_geom, wanted_url = geom(self.wanted_map)
        if not wanted_map_geom:
            # try to prepend us\ to the self.wanted_map
            wanted_map_geom, wanted_url = geom('us/'+self.wanted_map)
            if wanted_map_geom:
                self.wanted_map = 'us/'+self.wanted_map
            else:
                log.error(
                    'failed to find country or region %s in Geofabrik json file', self.wanted_map)
                sys.exit()

        # convert to shape (multipolygon)
        wanted_region = shape(wanted_map_geom)
        #print (f'shape = {wanted_region}')

        # get bounding box
        (bbox_left, bbox_bottom, bbox_right, bbox_top) = wanted_region.bounds

        # convert bounding box to list of tiles at zoom level 8
        (top_x, top_y) = deg2num(bbox_top, bbox_left)
        (bot_x, bot_y) = deg2num(bbox_bottom, bbox_right)

        # and stay within the allowed tilenumber range!
        if top_x < 0:
            top_x = 0
        if top_x > 255:
            top_x = 255
        if top_y < 0:
            top_y = 0
        if top_y > 255:
            top_y = 255
        if bot_x < 0:
            bot_x = 0
        if bot_x > 255:
            bot_x = 255
        if bot_y < 0:
            bot_y = 0
        if bot_y > 255:
            bot_y = 255

        # Build list of tiles from the bounding box
        bbox_tiles = []
        for x_value in range(top_x, bot_x + 1):
            for y_value in range(top_y, bot_y + 1):
                (tile_top, tile_left) = num2deg(x_value, y_value)
                (tile_bottom, tile_right) = num2deg(x_value+1, y_value+1)
                if tile_left < -180:
                    tile_left = -180
                if tile_left > 180:
                    tile_left = 180
                if tile_right < -180:
                    tile_right = -180
                if tile_right > 180:
                    tile_right = 180
                if tile_top < -90:
                    tile_top = -90
                if tile_top > 90:
                    tile_top = 90
                if tile_bottom < -90:
                    tile_bottom = -90
                if tile_bottom > 90:
                    tile_bottom = 90
                bbox_tiles.append({'x': x_value, 'y': y_value, 'tile_left': tile_left,
                                   'tile_top': tile_top, 'tile_right': tile_right,
                                   'tile_bottom': tile_bottom})

        log.info('Searching for needed maps, this can take a while.')
        tiles_of_input = find_needed_countries(
            bbox_tiles, self.wanted_map, wanted_region)
        #print (f'Country= {country}')

        return tiles_of_input


def deg2num(lat_deg, lon_deg, zoom=8):
    """
    Convert on./lat. to tile numbers
    """
    lat_rad = math.radians(lat_deg)
    n = 2.0 ** zoom
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    ytile = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)
    return (xtile, ytile)


def num2deg(xtile, ytile, zoom=8):
    """
    Convert tile numbers to lon./lat.
    """
    n = 2.0 ** zoom
    lon_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return (lat_deg, lon_deg)


def geom(wanted):
    """
    Get the Geofabrik outline of the desired country/region from the Geofabrik json file
    and the download url of the map.
    input parameter is the name of the desired country/region as use by Geofabric
    in their json file.
    """
    with open(GEOFABRIK_PATH, encoding='utf8') as file_handle:
        data = geojson.load(file_handle)
    file_handle.close()

    # loop through all entries in the json file to find the one we want
    for feature in data.features:
        props = feature.properties
        ident_no = props.get('id', '')
        if ident_no != wanted:
            continue
        #print (props.get('urls', ''))
        wurls = props.get('urls', '')
        return (feature.geometry, wurls.get('pbf', ''))
    return None, None


def find_geofbrik_parent(name, geofabrik_json):
    """
    Get the parent map/region of a region from the already loaded json data
    """
    for feature in geofabrik_json.features:
        props = feature.properties
        ident_no = props.get('id', '')
        if ident_no != name:
            continue
        return (props.get('parent', ''), props.get('id', ''))
    return None, None


def find_geofbrik_url(name, geofabrik_json):
    """
    Get the map download url from a region with the already loaded json data
    """
    for feature in geofabrik_json.features:
        props = feature.properties
        ident_no = props.get('id', '')
        if ident_no != name:
            continue
        #print (props.get('urls', ''))
        wurls = props.get('urls', '')
        return wurls.get('pbf', '')
    return None


def find_needed_countries(bbox_tiles, wanted_map, wanted_region_polygon):
    """
    Find the maps to download from Geofabrik for a given range of tiles
    arguments are
      - list of tiles of the desired region bounding box
      - name of desired region as used in Geofabrik json file
      - polygon of desired region as present in the Geofabrik json file
    """
    output = []

    with open(GEOFABRIK_PATH, encoding='utf8') as file_handle:
        geofabrik_json_data = geojson.load(file_handle)
    file_handle.close()

    # itterate through tiles and find Geofabrik regions that are in the tiles
    counter = 1
    for tile in bbox_tiles:
        # Do progress indicator every 50 tiles
        if counter % 50 == 0:
            log.info(
                'Processing tile %s of %s', counter, len(bbox_tiles)+1)
        counter += 1

        parent_added = 0
        force_added = 0

        # example contents of tile: {'index': 0, 'x': 130, 'y': 84, 'tile_left': 2.8125,
        # 'tile_top': 52.48278022207821, 'tile_right': 4.21875, 'tile_bottom': 51.6180165487737}
        # convert tile x/y to tile polygon lon/lat
        poly = Polygon([(tile["tile_left"], tile["tile_top"]), (tile["tile_right"],
                                                                tile["tile_top"]), (tile["tile_right"], tile["tile_bottom"]),
                        (tile["tile_left"], tile["tile_bottom"]), (tile["tile_left"],
                                                                   tile["tile_top"])])

        # (re)initialize list of needed maps and their url's
        must_download_maps = []
        must_download_urls = []

        # itterate through countries/regions in the geofabrik json file
        for regions in geofabrik_json_data.features:
            props = regions.properties
            parent = props.get('parent', '')
            regionname = props.get('id', '')
            rurls = props.get('urls', '')
            rurl = rurls.get('pbf', '')
            rgeom = regions.geometry
            rshape = shape(rgeom)

            #print (f'Processing region: {regionname}')

            # check if the region we are processing is needed for the tile we are processing

            # If currently processing country/region IS the desired country/region
            if regionname == wanted_map:
                # Check if it is part of the tile we are processing
                if rshape.intersects(poly):  # if so
                    # catch special_regions like (former) colonies where the map of the region is not fysically in the map of the parent country.
                    # example Guadeloupe, it's parent country is France but Guadeloupe is not located within the region covered by the map of France
                    if wanted_map not in special_regions:
                        # If we are proseccing a sub-region add the parent of this sub-region
                        # to the must download list.
                        # This to prevent downloading several small regions AND it's containing region
                        # we are processing a sub-regiongo find the parent region:
                        if parent not in geofabrik_regions and regionname not in geofabrik_regions:
                            # we are processing a sub-regiongo find the parent region
                            x_value = 0
                            # handle sub-sub-regions like unterfranken->bayern->germany
                            while parent not in geofabrik_regions:
                                parent, child = find_geofbrik_parent(
                                    parent, geofabrik_json_data)
                                if parent in geofabrik_regions:
                                    parent = child
                                    break
                                if x_value > 10:  # prevent endless loop
                                    log.error(
                                        'Can not find parent map of region: %s', regionname)
                                    sys.exit()
                                x_value += 1
                            if parent not in must_download_maps:
                                must_download_maps.append(parent)
                                must_download_urls.append(
                                    find_geofbrik_url(parent, geofabrik_json_data))
                                #parent_added = 1
                        else:
                            if regionname not in must_download_maps:
                                must_download_maps.append(regionname)
                                must_download_urls.append(rurl)
                    else:
                        # wanted_map is a special region like Guadeloupe, France
                        if regionname not in must_download_maps:
                            must_download_maps.append(regionname)
                            must_download_urls.append(rurl)
                    # if there is an intersect, force the tile to be put in the output
                    force_added = 1
                else:  # currently processing tile does not contain, a part of, the desired region
                    continue

            # currently processing country/region is NOT the desired country/region but might be
            # in the tile (neighbouring country)
            if regionname != wanted_map:
                # check if we are processing a country or a sub-region.
                # For countries only process other countries. also block special geofabrik sub regions
                if parent in geofabrik_regions and regionname not in block_download:
                    # processing a country and no special sub-region
                    # check if rshape is subset of desired region. If so discard it
                    if wanted_region_polygon.contains(rshape):
                        #print (f'\t{regionname} is a subset of {wanted_map}, discard it')
                        continue
                    # check if rshape is a superset of desired region. if so discard it
                    if rshape.contains(wanted_region_polygon):
                        #print (f'\t{regionname} is a superset of {wanted_map}, discard it')
                        # if regionname not in must_download_maps:
                        #    must_download_maps.append (regionname)
                        #    must_download_urls.append (rurl)
                        #    parent_added = 1
                        continue
                    # Check if rshape is a part of the tile
                    if rshape.intersects(poly):
                        #print(f'\tintersecting tile: {regionname} tile={tile}')
                        if regionname not in must_download_maps:
                            must_download_maps.append(regionname)
                            must_download_urls.append(rurl)

        # If this tile contains the desired region, add it to the output
        #print (f'map= {wanted_map}\tmust_download= {must_download_maps}\tparent_added= {parent_added}\tforce_added= {force_added}')
        if wanted_map in must_download_maps or parent_added == 1 or force_added == 1:
            # first replace any forward slashes with underscores (us/texas to us_texas)
            must_download_maps = [sub.replace(
                '/', '_') for sub in must_download_maps]
            output.append({'x': tile['x'], 'y': tile['y'], 'left': tile['tile_left'], 'top': tile['tile_top'],
                          'right': tile['tile_right'], 'bottom': tile['tile_bottom'], 'countries': must_download_maps, 'urls': must_download_urls})
    return output



"""
constants
"""
#!/usr/bin/python

import os
from pathlib import Path

# User
USER_WAHOO_MC = os.path.join(str(Path.home()), 'wahooMapsCreatorData')
USER_DL_DIR = os.path.join(USER_WAHOO_MC, '_download')
USER_MAPS_DIR = os.path.join(USER_DL_DIR, 'maps')
LAND_POLYGONS_PATH = os.path.join(
    USER_DL_DIR, 'land-polygons-split-4326', 'land_polygons.shp')
GEOFABRIK_PATH = os.path.join(USER_DL_DIR, 'geofabrik.json')
USER_OUTPUT_DIR = os.path.join(USER_WAHOO_MC, '_tiles')
USER_CONFIG_DIR = os.path.join(USER_WAHOO_MC, '_config')
USER_TOOLING_WIN_DIR = os.path.join(USER_DL_DIR, 'tooling_win')
OSMOSIS_WIN_FILE_PATH = os.path.join(
    USER_TOOLING_WIN_DIR, 'Osmosis', 'bin', 'osmosis.bat')

# Python Package - wahooMapsCreator directory
WAHOO_MC_DIR = os.path.dirname(__file__)
RESOURCES_DIR = os.path.join(WAHOO_MC_DIR, 'resources')
TOOLING_WIN_DIR = os.path.join(WAHOO_MC_DIR, 'tooling_win')
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
VERSION = '3.1.1'

Translate_Country = {
    'alaska':                           'united_states_of_america',
    'anguilla':                         'central-america',
    'bahrain':                          'gcc-states',
    'bosnia_and_herzegovina':           'bosnia-herzegovina',
    'british_virgin_islands':           'central-america',
    'british_indian_ocean_territory':   'asia',
    'brunei':                           'malaysia-singapore-brunei',
    'burkina_faso':                     'burkina-faso',
    'canary_islands':					'canary-islands',
    'cape_verde':                       'cape-verde',
    'christmas_island':                 'indonesia',
    'cocos_islands':                    'australia',
    'cote_d_ivoire':                    'ivory-coast',
    'czech_republic':                   'czech-republic',
    'democratic_republic_of_the_congo': 'congo-democratic-republic',
    'east_timor':                       'indonesia',
    'faroe_islands':                    'faroe-islands',
    'gambia':                           'senegal-and-gambia',
    'gibraltar':                        'spain',
    'guernsey':                         'guernsey-jersey',
    'hong_kong':                        'china',
    'ireland':                          'ireland-and-northern-ireland',
    'isle_of_man':                      'isle-of-man',
    'israel':                           'israel-and-palestine',
    'jersey':                           'guernsey-jersey',
    'kuwait':                           'gcc-states',
    'macao':                            'china',
    'malaysia':                         'malaysia-singapore-brunei',
    'north_korea':                      'north-korea',
    'oman':                             'gcc-states',
    'palestina':                        'israel-and-palestine',
    'papua_new_guinea':                 'asia',
    'paracel_islands':                  'china',
    'republic_of_congo':                'congo-brazzaville',
    'saint-martin':                     'central-america',
    'saint_helena':                     'saint-helena-ascension-and-tristan-da-cunha',
    'san_marino':                       'italy',
    'saudi_arabia':                     'gcc-states',
    'senegal':                          'senegal-and-gambia',
    'singapore':                        'malaysia-singapore-brunei',
    'sint_maarten':                     'central-america',
    'south_africa':                     'south-africa',
    'south_korea':                      'south-korea',
    'spratly_islands':                  'asia',
    'sri_lanka':                        'sri-lanka',
    'svalbard_and_jan_mayen':           'norway',
    'united_arab_emirates':             'gcc-states',
    'united_kingdom':                   'great-britain',
    'united_states_virgin_islands':     'central-america',
    'vatican_city':                     'italy',
    'virgin_islands_u.s.':              'central-america',
    'western_sahara':                   'morocco',
    'qatar':                            'gcc-states',
    'åland':                            'finland',
    'new_mexico':                       'new-mexico',
    'american_samoa':                   'samoa',
    'commonwealth_of_the_northern_mariana_islands':                   'american-oceania',
    'northern_mariana_islands':         'american-oceania',
    'new_york':                         'new-york',
    'new_hampshire':                    'new-hampshire',
    'new_jersey':                       'new-jersey',
    'rhode_island':                     'rhode-island',
    'district_of_columbia':             'district-of-columbia',
    'north_carolina':                   'north-carolina',
    'south_carolina':                   'south-carolina',
    'north_dakota':                     'north-dakota',
    'south_dakota':                     'south-dakota',
    'west_virginia':                    'west-virginia'
}

continents = ['europe', 'unitedstates', 'north-america', 'south-america', 'asia', 'oceania',
              'africa', 'antarctica']

africa = ['algeria', 'angola', 'benin', 'botswana', 'burkina_faso', 'burundi',
          'cameroon', 'canary_islands', 'cape_verde', 'central_african_republic', 'chad', 'comoros',
          'cote_d_ivoire', 'democratic_republic_of_the_congo', 'djibouti', 'egypt',
          'equatorial_guinea', 'eritrea', 'ethiopia', 'french_southern_territories', 'gabon',
          'gambia', 'ghana', 'guinea-bissau', 'guinea', 'kenya', 'lesotho', 'liberia',
          'libya', 'madagascar', 'malawi', 'mali', 'mauritania', 'mauritius', 'mayotte',
          'morocco', 'mozambique', 'namibia', 'niger', 'nigeria', 'republic_of_congo',
          'reunion', 'rwanda', 'saint_helena', 'sao_tome_and_principe', 'senegal',
          'seychelles', 'sierra_leone', 'somalia', 'south_africa', 'sudan', 'swaziland',
          'tanzania', 'togo', 'tunisia', 'uganda', 'western_sahara', 'zambia', 'zimbabwe']

antarctica = ['antarctica', 'bouvet_island', 'heard_island_and_mcdonald_islands',
              'south_georgia_and_the_south_sandwich_islands']

asia = ['afghanistan', 'armenia', 'azerbaijan', 'bahrain', 'bangladesh', 'bhutan',
        'british_indian_ocean_territory', 'brunei', 'cambodia', 'china', 'christmas_island',
        'cocos_islands', 'cyprus', 'east_timor', 'hong_kong', 'india', 'indonesia',
        'iran', 'iraq', 'israel', 'japan', 'jordan', 'kazakhstan', 'kuwait',
        'kyrgyzstan', 'laos', 'lebanon', 'macao', 'malaysia', 'maldives', 'mongolia',
        'myanmar', 'nepal', 'north_korea', 'oman', 'pakistan', 'palestina',
        'paracel_islands', 'philippines', 'qatar', 'saudi_arabia', 'singapore',
        'south_korea', 'spratly_islands', 'sri_lanka', 'syria', 'taiwan', 'tajikistan',
        'thailand', 'turkey', 'turkmenistan', 'united_arab_emirates', 'uzbekistan',
        'vietnam', 'yemen']

europe = ['albania', 'andorra', 'austria', 'belarus', 'belgium', 'bosnia_and_herzegovina',
          'bulgaria', 'croatia', 'czech_republic', 'denmark', 'estonia', 'faroe_islands',
          'finland', 'france', 'germany', 'gibraltar', 'greece', 'guernsey', 'hungary',
          'iceland', 'ireland', 'isle_of_man', 'italy', 'jersey', 'latvia', 'liechtenstein',
          'lithuania', 'luxembourg', 'macedonia', 'malta', 'moldova', 'monaco', 'montenegro',
          'netherlands', 'norway', 'poland', 'portugal', 'romania', 'russia', 'san_marino',
          'serbia', 'slovakia', 'slovenia', 'spain', 'svalbard_and_jan_mayen', 'sweden',
          'switzerland', 'ukraine', 'united_kingdom', 'vatican_city', 'åland']

northamerica = ['anguilla', 'antigua_and_barbuda', 'bahamas', 'barbados', 'belize',
                'bermuda', 'british_virgin_islands', 'canada', 'cayman_islands', 'costa_rica',
                'cuba', 'dominica', 'dominican_republic', 'el_salvador', 'greenland', 'grenada',
                'guadeloupe', 'guatemala', 'haiti', 'honduras', 'jamaica', 'martinique',
                'mexico', 'montserrat', 'nicaragua', 'panama', 'saint-barthélemy', 'saint-martin',
                'saint_kitts_and_nevis', 'saint_lucia', 'saint_vincent_and_the_grenadines',
                'sint_maarten', 'turks_and_caicos_islands', 'virgin_islands_u.s.']

oceania = ['australia', 'cook_islands', 'fiji', 'french_polynesia', 'kiribati',
           'marshall_islands', 'micronesia', 'nauru', 'new_caledonia', 'new_zealand', 'niue',
           'norfolk_island', 'northern_mariana_islands', 'palau', 'papua_new_guinea',
           'pitcairn_islands', 'samoa', 'solomon_islands', 'tokelau', 'tonga', 'tuvalu',
           'united_states_minor_outlying_islands', 'vanuatu', 'wallis_and_futuna']

southamerica = ['argentina', 'aruba', 'bolivia', 'bonaire_saint_eustatius_and_saba', 'brazil',
                'chile', 'clipperton_island', 'colombia', 'curacao', 'ecuador',
                'falkland_islands', 'french_guiana', 'guyana', 'paraguay', 'peru',
                'saint_pierre_and_miquelon', 'suriname', 'trinidad_and_tobago',
                'uruguay', 'venezuela']

unitedstates = ['alabama', 'alaska', 'american_samoa', 'arizona', 'arkansas',
                'california', 'colorado', 'commonwealth_of_the_northern_mariana_islands', 'connecticut',
                'delaware', 'district_of_columbia', 'florida', 'georgia', 'guam', 'hawaii',
                'idaho', 'illinois', 'indiana', 'iowa', 'kansas', 'kentucky', 'louisiana',
                'maine', 'maryland', 'massachusetts', 'michigan', 'minnesota', 'mississippi',
                'missouri', 'montana', 'nebraska', 'nevada', 'new_hampshire', 'new_jersey',
                'new_mexico', 'new_york', 'north_carolina', 'north_dakota', 'ohio', 'oklahoma',
                'oregon', 'pennsylvania', 'puerto_rico', 'rhode_island', 'south_carolina',
                'south_dakota', 'tennessee', 'texas', 'united_states_virgin_islands', 'utah',
                'vermont', 'virginia', 'washington', 'west_virginia', 'wisconsin', 'wyoming']

africa_geofabrik = ['algeria', 'angola', 'benin', 'botswana', 'burkina-faso', 'burundi',
                    'cameroon', 'canary-islands', 'cape-verde', 'central african republic', 'chad',
                    'comores', 'Congo (Republic/Brazzaville)', 'congo-democratic-republic', 'djibouti',
                    'egypt', 'Equatorial Guinea', 'eritrea', 'ethiopia', 'gabon', 'ghana', 'guinea',
                    'guinea-bissau', 'ivory-coast', 'kenya', 'lesotho', 'liberia', 'libya', 'madagascar',
                    'malawi', 'mali', 'mauritania', 'mauritius', 'morocco', 'mozambique', 'namibia',
                    'niger', 'nigeria', 'rwanda', 'saint-helena-ascension-and-tristan-da-cunha',
                    'Sao Tome and Principe', 'Senegal and Gambia', 'seychelles', 'Sierra Leone',
                    'somalia', 'south-africa', 'South Sudan', 'sudan', 'swaziland', 'tanzania', 'togo',
                    'tunisia', 'uganda', 'zambia', 'zimbabwe']

antarctica_geofabrik = ['antarctica']

asia_geofabrik = ['afghanistan', 'armenia', 'azerbaijan', 'bangladesh', 'bhutan', 'cambodia',
                  'china', 'gcc-states', 'india', 'indonesia', 'iran', 'iraq', 'israel-and-palestine',
                  'japan', 'jordan', 'kazakhstan', 'kyrgyzstan', 'laos', 'lebanon',
                  'malaysia-singapore-brunei', 'maldives', 'mongolia', 'myanmar', 'nepal', 'north-korea',
                  'pakistan', 'philippines', 'russian federation', 'south-korea', 'sri-lanka', 'syria',
                  'taiwan', 'tajikistan', 'thailand', 'turkmenistan', 'uzbekistan', 'vietnam', 'yemen']

australiaoceania_geofabrik = ['american-oceania', 'australia', 'cook islands', 'fiji',
                              'île de clipperton', 'kiribati', 'marshall islands', 'micronesia', 'nauru',
                              'new caledonia', 'new zealand', 'niue', 'palau', 'papua new guinea',
                              'pitcairn islands', 'polynesie-francaise', 'samoa', 'solomon islands', 'tokelau',
                              'tonga', 'tuvalu', 'vanuatu', 'wallis et futuna']

centralamerica_geofabrik = ['bahamas', 'belize', 'costa rica', 'cuba', 'el salvador', 'guatemala',
                            'haiti and dominican republic', 'honduras', 'jamaica', 'nicaragua']

europe_geofabrik = ['albania', 'andorra', 'austria', 'azores', 'belarus', 'belgium',
                    'bosnia-herzegovina', 'bulgaria', 'croatia', 'cyprus', 'czech-republic', 'denmark',
                    'estonia', 'faroe-islands', 'finland', 'france', 'georgia', 'germany', 'great-britain',
                    'greece', 'guernsey-jersey', 'hungary', 'iceland', 'ireland-and-northern-ireland',
                    'isle-of-man', 'italy', 'kosovo', 'latvia', 'liechtenstein', 'lithuania', 'luxembourg',
                    'macedonia', 'malta', 'moldova', 'monaco', 'montenegro', 'netherlands', 'norway',
                    'poland', 'portugal', 'romania', 'russian federation', 'serbia', 'slovakia',
                    'slovenia', 'spain', 'sweden', 'switzerland', 'turkey', 'ukraine']

northamerica_geofabrik = ['canada', 'greenland', 'mexico', 'us midwest', 'us northeast',
                          'us pacific', 'us south', 'us west']

northamerica_us_geofabrik = ['alabama', 'alaska', 'arizona', 'arkansas', 'california', 'colorado',
                             'connecticut', 'delaware', 'district-of-columbia', 'florida',
                             'georgia', 'hawaii', 'idaho', 'illinois', 'indiana', 'iowa', 'kansas',
                             'kentucky', 'louisiana', 'maine', 'maryland', 'massachusetts',
                             'michigan', 'minnesota', 'mississippi', 'missouri', 'montana',
                             'nebraska', 'nevada', 'new-hampshire', 'new-jersey', 'new-mexico',
                             'new-york', 'north-carolina', 'north-dakota', 'ohio', 'oklahoma'
                             'oregon', 'pennsylvania', 'puerto-rico', 'rhode-island',
                             'south-carolina', 'south-dakota', 'tennessee', 'texas',
                             'us-virgin-islands', 'utah', 'vermont', 'virginia', 'washington',
                             'west-virginia', 'wisconsin', 'wyoming']

southamerica_geofabrik = ['argentina', 'bolivia', 'brazil', 'chile', 'colombia', 'ecuador',
                          'paraguay', 'peru', 'suriname', 'uruguay', 'venezuela']

germany_subregions_geofabrik = ['baden-württemberg', 'bayern', 'berlin',
                                'brandenburg (mit berlin)', 'bremen', 'hamburg', 'hessen', 'mecklenburg-vorpommern',
                                'niedersachsen', 'nordrhein-westfalen', 'rheinland-pfalz', 'saarland', 'sachsen',
                                'sachsen-anhalt', 'schleswig-holstein', 'thüringen']

france_subregions_geofabrik = ['alsace', 'aquitaine', 'auvergne', 'basse-normandie', 'bourgogne',
                               'bretagne', 'centre', 'champagne ardenne', 'corse', 'franche comte', 'guadeloupe',
                               'guyane', 'haute-normandie', 'ile-de-france', 'languedoc-roussillon', 'limousin',
                               'lorraine', 'martinique', 'mayotte', 'midi-pyrenees', 'nord-pas-de-calais',
                               'pays de la loire', 'picardie', 'poitou-charentes', 'provence alpes-cote-d\'azur',
                               'reunion', 'rhone-alpes']
# great-britain_subregions_geofabrik = ['england', 'scotland', 'wales']

italy_subregions_geofabrik = [
    'Centro', 'Isole', 'Nord-Est', 'Nord-Ovest', 'Sud']

noregion_geofabrik = ['russia', 'asia']

geofabrik_regions = ['africa', 'antarctica', 'asia', 'australia-oceania',
                     'central-america', 'europe', 'north-america', 'south-america']

block_download = ['dach', 'alps', 'britain-and-ireland', 'south-africa-and-lesotho',
                  'us-midwest', 'us-northeast', 'us-pacific', 'us-south', 'us-west']

# Special_regions like (former) colonies where the map of the wanted region is not present in the map of the parent country.
# example Guadeloupe, it's Geofabrik parent country is France but Guadeloupe is not located within the region covered by the map of France.
special_regions = ['guadeloupe', 'guyane', 'martinique', 'mayotte', 'reunion']



"""
functions and object for checking for up-to-date & downloading land_polygons file and OSM maps
"""
#!/usr/bin/python

# import official python packages
import glob
import os
import os.path
import sys
import time
import logging
import platform

# import custom python packages
from wahoomc.file_directory_functions import download_url_to_file, unzip
from wahoomc.constants_functions import translate_country_input_to_geofabrik, \
    get_geofabrik_region_of_country, get_tooling_win_path

from wahoomc.constants import USER_DL_DIR
from wahoomc.constants import USER_MAPS_DIR
from wahoomc.constants import LAND_POLYGONS_PATH
from wahoomc.constants import GEOFABRIK_PATH
from wahoomc.constants import OSMOSIS_WIN_FILE_PATH
from wahoomc.constants import USER_TOOLING_WIN_DIR

log = logging.getLogger('main-logger')


def older_than_x_days(file_creation_timestamp, max_days_old):
    """
    check if given timestamp is older that given days
    """
    to_old_timestamp = time.time() - 60 * 60 * 24 * max_days_old

    return bool(file_creation_timestamp < to_old_timestamp)


def download_file(target_filepath, url, target_dir=""):
    """
    download given file and eventually unzip it
    """
    logging_filename = target_filepath.split(os.sep)[-1]
    log.info('-' * 80)
    log.info('# Downloading %s file', logging_filename)
    if url.split('.')[-1] == 'zip':
        # build target-filepath based on last element of URL
        last_part = url.rsplit('/', 1)[-1]
        dl_file_path = os.path.join(USER_DL_DIR, last_part)
        # download URL to file
        download_url_to_file(url, dl_file_path)

        # if a target directory is given --> extract into that folder
        if target_dir:
            target_path = target_dir
        else:
            target_path = USER_DL_DIR

        # unpack it
        unzip(dl_file_path, target_path)

        os.remove(dl_file_path)
    else:
        # no zipping --> directly download to given target filepath
        download_url_to_file(url, target_filepath)
    # Check if file exists (if target file exists)
    if not os.path.isfile(target_filepath):
        log.error('! failed to find %s', target_filepath)
        sys.exit()
    else:
        log.info('+ Downloaded: %s', target_filepath)


def get_osm_pbf_filepath_url(country):
    """
    build the geofabrik download url to a countries' OSM file and download filepath
    """

    transl_c = translate_country_input_to_geofabrik(country)
    region = get_geofabrik_region_of_country(country)
    if region != 'no':
        url = 'https://download.geofabrik.de/' + region + \
            '/' + transl_c + '-latest.osm.pbf'
    else:
        url = 'https://download.geofabrik.de/' + transl_c + '-latest.osm.pbf'

    map_file_path = os.path.join(
        USER_MAPS_DIR, f'{transl_c}' + '-latest.osm.pbf')

    # return URL and download filepath
    return map_file_path, url


def download_tooling_win():
    """
    check for Windows tooling and download if not here already
    this is done to bring down the filesize of the python module
    """
    # Windows
    if platform.system() == "Windows":
        if not os.path.isfile(OSMOSIS_WIN_FILE_PATH):
            log.info('# Need to download Osmosis application for Windows')
            download_file(OSMOSIS_WIN_FILE_PATH,
                          'https://github.com/openstreetmap/osmosis/releases/download/0.48.3/osmosis-0.48.3.zip',
                          get_tooling_win_path('Osmosis', in_user_dir=True))

        new_p = os.path.join(USER_TOOLING_WIN_DIR,
                             'Osmosis', 'lib', 'default', 'mapsforge-map-writer-0.18.0-jar-with-dependencies.jar')
        if not os.path.isfile(new_p):
            log.info('# Need to download Osmosis mapwriter plugin for Windows')
            download_file(new_p,
                          'https://search.maven.org/remotecontent?filepath=org/mapsforge/mapsforge-map-writer/0.18.0/mapsforge-map-writer-0.18.0-jar-with-dependencies.jar')

        if not os.path.isfile(get_tooling_win_path('osmfilter.exe', in_user_dir=True)):
            log.info('# Need to download osmfilter application for Windows')

            download_file(get_tooling_win_path('osmfilter.exe', in_user_dir=True),
                          'http://m.m.i24.cc/osmfilter.exe')


class Downloader:
    """
    This is the class to check and download maps / artifacts"
    """

    def __init__(self, max_days_old, force_download, tiles_from_json=None, border_countries=None):
        self.max_days_old = max_days_old
        self.force_download = force_download
        self.tiles_from_json = tiles_from_json
        self.border_countries = border_countries

        self.need_to_dl = []

    def should_geofabrik_file_be_downloaded(self):
        """
        check geofabrik file if not existing or is not up-to-date

        # if geofabrik file needs to be downloaded, force-processing is set because there might be
        # a change in the geofabrik file
        """
        if self.check_file(GEOFABRIK_PATH) is True or \
                self.force_download is True:
            log.info('# Need to download geofabrik file')

            return True

        return False

    def download_geofabrik_file(self):
        """
        download geofabrik file
        """
        download_file(GEOFABRIK_PATH,
                      'https://download.geofabrik.de/index-v1.json')

        log.info('+ download geofabrik.json file: OK')

    def check_land_polygons_file(self):
        """
        check land_polygons file if not existing or are not up-to-date
        """

        if self.check_file(LAND_POLYGONS_PATH) is True or \
                self.force_download is True:
            log.info('+ Need to download land polygons file')
            self.need_to_dl.append('land_polygons')

    def download_files_if_needed(self):
        """
        check land_polygons and OSM map files if not existing or are not up-to-date
        and download if needed
        """
        if 'land_polygons' in self.need_to_dl:
            download_file(LAND_POLYGONS_PATH,
                          'https://osmdata.openstreetmap.de/download/land-polygons-split-4326.zip')

        # log.info('+ download land_polygons.shp file: OK')

        if 'osm_pbf' in self.need_to_dl:
            self.download_osm_pbf_file()

    def check_file(self, target_filepath):
        """
        check if given file is up-to-date
        """

        need_to_download = False
        logging_filename = target_filepath.rsplit('/', 1)[-1]

        log.info('-' * 80)
        log.info('# check %s file', logging_filename)

        # Check for expired file and delete it
        try:
            if self.should_file_be_downloaded(target_filepath):
                log.info('+ Deleting old %s file', logging_filename)
                os.remove(target_filepath)
                need_to_download = True

        except FileNotFoundError:
            need_to_download = True

        # if file does not exists --> download
        if not os.path.exists(target_filepath) or \
                not os.path.isfile(target_filepath):
            need_to_download = True

            log.info('+ %s file needs to be downloaded', logging_filename)

        return need_to_download

    def check_osm_pbf_file(self):
        """
        check if the relevant countries' OSM files are up-to-date
        """

        log.info('-' * 80)
        log.info('# check countries .osm.pbf files')

        # Check for expired maps and delete them
        log.info('+ Checking for old maps and remove them')

        for country in self.border_countries:
            # get translated country (geofabrik) of country
            # do not download the same file for different countries
            # --> e.g. China, Hong Kong and Macao, see Issue #11
            transl_c = translate_country_input_to_geofabrik(country)

            # check for already existing .osm.pbf file
            map_file_path = glob.glob(
                f'{USER_MAPS_DIR}/{transl_c}-latest.osm.pbf')
            if len(map_file_path) != 1:
                map_file_path = glob.glob(
                    f'{USER_MAPS_DIR}/**/{transl_c}-latest.osm.pbf')

            # delete .osm.pbf file if out of date
            if len(map_file_path) == 1 and os.path.isfile(map_file_path[0]):
                if self.should_file_be_downloaded(map_file_path[0]):
                    log.info(
                        '+ mapfile for %s: deleted. Input: %s.', transl_c, country)
                    os.remove(map_file_path[0])
                    self.need_to_dl.append('osm_pbf')
                else:
                    self.border_countries[country] = {
                        'map_file': map_file_path[0]}
                    log.info(
                        '+ mapfile for %s: up-to-date. Input: %s.', transl_c, country)

            # mark country .osm.pbf file for download if there exists no file or it is no file
            map_file_path = self.border_countries[country].get('map_file')
            if map_file_path is None or (not os.path.isfile(map_file_path) or self.force_download):
                self.border_countries[country]['download'] = True
                self.need_to_dl.append('osm_pbf')

    def download_osm_pbf_file(self):
        """
        download countries' OSM files
        """
        for country, item in self.border_countries.items():
            try:
                if item['download'] is True:
                    map_file_path, url = get_osm_pbf_filepath_url(country)
                    download_file(map_file_path, url)
                    self.border_countries[country] = {
                        'map_file': map_file_path}
            except KeyError:
                pass

    def should_file_be_downloaded(self, file_path):
        """
        check if given file should be downloaded
        - older that max_days old OR
        - force_download is set
        """

        chg_time = os.path.getmtime(file_path)

        return bool(older_than_x_days(chg_time, self.max_days_old) or self.force_download is True)
